from enum import auto, Enum


class VelocityTimeUnit(Enum):
    HOUR = auto()
    DAY = auto()
    WEEK = auto()
    MONTH = auto()


class HealthStatus(Enum):
    GREEN = auto()
    YELLOW = auto()
    ORANGE = auto()
    RED = auto()


class SeniorityLevel(Enum):
    JUNIOR = auto()
    MIDDLE = auto()
    SENIOR = auto()



import datetime
import unittest

from sd_metrics_lib.sources.azure.query import AzureSearchQueryBuilder


class AzureSearchQueryBuilderTestCase(unittest.TestCase):

    def test_empty_builder_returns_select_statement(self):
        # given
        builder = AzureSearchQueryBuilder()
        # when
        query = builder.build_query()
        # then
        self.assertEqual('SELECT [System.Id] FROM workitems', query)

    def test_task_ids_only(self):
        # given
        builder = AzureSearchQueryBuilder(task_ids=[1, 2, 3])
        # when
        query = builder.build_query()
        # then
        expected = "SELECT [System.Id] FROM workitems WHERE [System.Id] IN (1, 2, 3)"
        self.assertEqual(expected, query)

    def test_full_query_build_with_multiple_teams_and_order_by(self):
        # given
        builder = AzureSearchQueryBuilder(
            projects=['Proj1', 'Proj2'],
            statuses=['Active', 'Closed'],
            task_types=['User Story', 'Bug'],
            teams=['Team A', 'Team B'],
            resolution_dates=(datetime.date(2022, 3, 1), datetime.date(2022, 3, 31)),
            last_modified_dates=(datetime.date(2022, 2, 1), None),
            raw_queries=['[Custom.Field] = 1', '   ', None],
            order_by='[System.ChangedDate] DESC'
        )
        # when
        query = builder.build_query()
        # then
        expected = f"SELECT [System.Id] FROM workitems WHERE {(
            "[System.TeamProject] IN ('Proj1', 'Proj2')"
            " AND [System.State] IN ('Active', 'Closed')"
            " AND [Microsoft.VSTS.Common.ClosedDate] >= '2022-03-01' AND [Microsoft.VSTS.Common.ClosedDate] <= '2022-03-31'"
            " AND [System.WorkItemType] IN ('User Story', 'Bug')"
            " AND ([System.AreaPath] UNDER 'Team A' OR [System.AreaPath] UNDER 'Team B')"
            " AND [System.ChangedDate] >= '2022-02-01'"
            " AND [Custom.Field] = 1"
        )} ORDER BY [System.ChangedDate] DESC"
        self.assertEqual(expected, query)

    def test_single_team_no_parentheses(self):
        # given
        builder = AzureSearchQueryBuilder(
            teams=['Team A']
        )
        # when
        query = builder.build_query()
        # then
        expected = "SELECT [System.Id] FROM workitems WHERE [System.AreaPath] UNDER 'Team A'"
        self.assertEqual(expected, query)

    def test_order_by_only_appended(self):
        # given
        builder = AzureSearchQueryBuilder(order_by='[System.CreatedDate] ASC')
        # when
        query = builder.build_query()
        # then
        self.assertEqual('SELECT [System.Id] FROM workitems ORDER BY [System.CreatedDate] ASC', query)


if __name__ == '__main__':
    unittest.main()



from typing import Dict

from sd_metrics_lib.sources.story_points import StoryPointExtractor


class JiraCustomFieldStoryPointExtractor(StoryPointExtractor):

    def __init__(self, custom_field_name, default_story_points_value=None) -> None:
        self.custom_field_name = custom_field_name
        self.default_value = default_story_points_value

    def get_story_points(self, task) -> float | None:
        story_points = self._extract_field_value(task)

        if isinstance(story_points, int) or isinstance(story_points, float):
            return story_points
        if story_points is None:
            return self.default_value
        if story_points.isdigit():
            return int(story_points)

        return self.default_value

    def _extract_field_value(self, task):
        try:
            return task['fields'][self.custom_field_name]
        except:
            return None


class JiraTShirtStoryPointExtractor(JiraCustomFieldStoryPointExtractor):

    def __init__(self, custom_field_name, story_point_mapping: Dict[str, int], default_story_points_value=0) -> None:
        super().__init__(custom_field_name, default_story_points_value)
        self.story_point_mapping = story_point_mapping

    def get_story_points(self, task) -> float | None:
        story_points = self._extract_field_value(task)
        if isinstance(story_points, int) or isinstance(story_points, float):
            return story_points
        if story_points is None:
            return self.default_value
        if story_points.isdigit():
            return int(story_points)

        story_points_lower = story_points.lower()
        return self.story_point_mapping.get(story_points_lower, self.default_value)



import concurrent
import math
from concurrent.futures import ThreadPoolExecutor, wait
from typing import Iterable

from sd_metrics_lib.sources.tasks import TaskProvider


class JiraTaskProvider(TaskProvider):

    def __init__(self,
                 jira_client,
                 query: str,
                 additional_fields: Iterable[str] = None,
                 thread_pool_executor: ThreadPoolExecutor = None) -> None:
        self.jira_client = jira_client
        self.query = query.strip()
        self.additional_fields = additional_fields
        if additional_fields is None:
            self._expand_str = None
        else:
            # For Jira, additional_fields correspond to expand values (e.g., 'changelog')
            self._expand_str = ",".join(self.additional_fields)
        self.thread_pool_executor = thread_pool_executor

    def get_tasks(self):
        first_page = self.jira_client.jql(self.query, expand=self._expand_str, limit=self._get_task_fetch_amount())
        first_page_tasks = first_page.get("issues", [])
        tasks_total_count = first_page.get("total", len(first_page_tasks))
        page_len = len(first_page_tasks)

        if tasks_total_count == 0 or page_len == 0:
            return []

        tasks = []
        tasks.extend(first_page_tasks)

        if page_len < tasks_total_count:
            amount_of_fetches = math.ceil(tasks_total_count / float(page_len))

            if self.thread_pool_executor is None:
                self._fetch_task_sync(tasks, amount_of_fetches, page_len)
            else:
                self._fetch_task_concurrently(tasks, amount_of_fetches, page_len)

        return tasks

    def _fetch_task_concurrently(self, tasks, amount_of_fetches, page_len):
        features = []
        for i in range(1, amount_of_fetches):
            next_search_start = i * page_len
            feature = self.thread_pool_executor.submit(self.jira_client.jql,
                                                       self.query,
                                                       expand=self._expand_str,
                                                       limit=self._get_task_fetch_amount(),
                                                       start=next_search_start)
            features.append(feature)
        done, not_done = wait(features, return_when=concurrent.futures.ALL_COMPLETED)
        for feature in done:
            tasks.extend(feature.result().get("issues", []))

    def _fetch_task_sync(self, tasks, amount_of_fetches, page_len):
        for i in range(1, amount_of_fetches):
            start = i * page_len
            current_page_result = self.jira_client.jql(self.query,
                                                       expand=self._expand_str,
                                                       limit=self._get_task_fetch_amount(),
                                                       start=start)
            current_page_tasks = current_page_result.get("issues", [])
            tasks.extend(current_page_tasks)

    def _get_task_fetch_amount(self):
        if self.thread_pool_executor is None:
            return 100
        else:
            return 50



class TShirtMapping:
    __MAPPING = {
        'xs': 1,
        's': 5,
        'm': 8,
        'l': 13,
        'xl': 21
    }

    @staticmethod
    def convert_into_points(size: str) -> int:
        size_code = size.lower()
        return TShirtMapping.__MAPPING.get(size_code, 0)

    @staticmethod
    def convert_into_size(story_point: int) -> str:
        if story_point <= 1:
            return 'XS'
        elif story_point <= 5:
            return 'S'
        elif story_point <= 8:
            return 'M'
        elif story_point <= 13:
            return 'L'
        elif story_point <= 21:
            return 'XL'
        return 'XXL'



from abc import ABC, abstractmethod
from typing import Dict, Callable, Optional

from sd_metrics_lib.utils.attributes import get_attribute_by_path


class WorklogExtractor(ABC):

    @abstractmethod
    def get_work_time_per_user(self, task) -> Dict[str, int]:
        pass


class TaskTotalSpentTimeExtractor(ABC):

    @abstractmethod
    def get_total_spent_time(self, task) -> int:
        pass


class ChainedWorklogExtractor(WorklogExtractor):

    def __init__(self, worklog_extractor_list: list[WorklogExtractor]) -> None:
        self.worklog_extractor_list = worklog_extractor_list

    def get_work_time_per_user(self, task):
        for worklog_extractor in self.worklog_extractor_list:
            work_time = worklog_extractor.get_work_time_per_user(task)
            if work_time is not None and len(work_time.keys()) != 0:
                return work_time
        return {}


class FunctionWorklogExtractor(WorklogExtractor):

    def __init__(self, func: Callable[[object], Optional[Dict[str, int]]]):
        self.func = func

    def get_work_time_per_user(self, task) -> Dict[str, int]:
        result = self.func(task)
        try:
            return {str(k): int(v) for k, v in (result or {}).items()}
        except Exception:
            return {}


class FunctionTotalSpentTimeExtractor(TaskTotalSpentTimeExtractor):

    def __init__(self, func: Callable[[object], Optional[int]]):
        self.func = func

    def get_total_spent_time(self, task) -> int:
        result = self.func(task)
        try:
            return int(result) if result is not None else 0
        except Exception:
            return 0


class AttributePathWorklogExtractor(WorklogExtractor):

    def __init__(self, attr_path: str):
        self._path = attr_path

    def get_work_time_per_user(self, task) -> Dict[str, int]:
        value = get_attribute_by_path(task, self._path, {})
        if isinstance(value, dict):
            try:
                return {str(k): int(v) for k, v in value.items()}
            except Exception:
                return {}
        return {}


class AttributePathTotalSpentTimeExtractor(TaskTotalSpentTimeExtractor):

    def __init__(self, attr_path: str, default: int = 0):
        self._path = attr_path
        self._default = default

    def get_total_spent_time(self, task) -> int:
        value = get_attribute_by_path(task, self._path, self._default)
        try:
            return int(value) if value is not None else self._default
        except Exception:
            return self._default



from datetime import datetime
from typing import Optional, List

from sd_metrics_lib.sources.abstract_worklog import AbstractStatusChangeWorklogExtractor
from sd_metrics_lib.sources.worklog import TaskTotalSpentTimeExtractor
from sd_metrics_lib.utils.worktime import WorkTimeExtractor, SimpleWorkTimeExtractor


class AzureStatusChangeWorklogExtractor(AbstractStatusChangeWorklogExtractor):

    def __init__(self,
                 transition_statuses: Optional[List[str]] = None,
                 user_filter: Optional[List[str]] = None,
                 time_format='%Y-%m-%dT%H:%M:%S.%f%z',
                 use_user_name: bool = True,
                 worktime_extractor: WorkTimeExtractor = SimpleWorkTimeExtractor()) -> None:
        super().__init__(transition_statuses=transition_statuses,
                         user_filter=user_filter,
                         worktime_extractor=worktime_extractor)
        self.time_format = time_format
        self.use_user_name = use_user_name

    def _extract_chronological_changes_sequence(self, task):
        fields = task.fields
        return fields.get('CustomExpand.WorkItemUpdate') or []

    def _is_user_change_entry(self, changelog_entry) -> bool:
        fields = changelog_entry.fields
        return fields and 'System.AssignedTo' in fields and fields['System.AssignedTo'].new_value is not None

    def _is_status_change_entry(self, changelog_entry) -> bool:
        fields = changelog_entry.fields
        return fields and 'System.State' in fields and fields['System.State'].new_value is not None

    def _extract_user_from_change(self, changelog_entry) -> str:
        return changelog_entry.fields['System.AssignedTo'].new_value['uniqueName']

    def _extract_change_time(self, changelog_entry):
        date_string = changelog_entry.fields['System.ChangedDate'].new_value
        try:
            return datetime.strptime(date_string, self.time_format)
        except ValueError:
            # Sometimes Azure API returns time without milliseconds
            return datetime.strptime(date_string, '%Y-%m-%dT%H:%M:%S%z')

    def _is_status_changed_into_required(self, changelog_entry) -> bool:
        if self.transition_statuses is None:
            return True
        return changelog_entry.fields['System.State'].new_value in self.transition_statuses

    def _is_status_changed_from_required(self, changelog_entry) -> bool:
        if self.transition_statuses is None:
            return True
        return changelog_entry.fields['System.State'].old_value in self.transition_statuses

    def _is_current_status_a_required_status(self, task) -> bool:
        if self.transition_statuses is None:
            return True
        fields = task.get('fields', {}) if isinstance(task, dict) else {}
        current = fields.get('System.State')
        return current in self.transition_statuses


class AzureTaskTotalSpentTimeExtractor(TaskTotalSpentTimeExtractor):

    def __init__(self, time_format='%Y-%m-%dT%H:%M:%S.%f%z') -> None:
        self.time_format = time_format

    def get_total_spent_time(self, task) -> int:
        resolution_date_str = task.fields['Microsoft.VSTS.Common.ClosedDate']
        if resolution_date_str is None:
            return 0

        resolution_date = self._convert_to_time(resolution_date_str)
        creation_date = self._convert_to_time(task.fields['System.CreatedDate'])
        spent_time = (resolution_date - creation_date)
        return int(spent_time.total_seconds())

    def _convert_to_time(self, date_string: str) -> datetime:
        try:
            return datetime.strptime(date_string, self.time_format)
        except ValueError:
            # Sometimes Azure API returns time without milliseconds
            return datetime.strptime(date_string, '%Y-%m-%dT%H:%M:%S%z')



from abc import ABC, abstractmethod
from typing import Callable, Optional

from sd_metrics_lib.utils.attributes import get_attribute_by_path


class StoryPointExtractor(ABC):

    @abstractmethod
    def get_story_points(self, task) -> float | None:
        pass


class ConstantStoryPointExtractor(StoryPointExtractor):

    def __init__(self, story_point_amount=1):
        self.value = story_point_amount

    def get_story_points(self, task) -> float | None:
        return self.value


class FunctionStoryPointExtractor(StoryPointExtractor):

    def __init__(self, func: Callable[[object], Optional[float]]):
        self.func = func

    def get_story_points(self, task) -> float | None:
        return self.func(task)


class AttributePathStoryPointExtractor(StoryPointExtractor):

    def __init__(self, attr_path: str, default: Optional[float] = None):
        self._path = attr_path
        self._default = default

    def get_story_points(self, task) -> float | None:
        value = get_attribute_by_path(task, self._path, self._default)
        if value is None:
            return None
        try:
            return float(value)
        except Exception:
            return self._default



import datetime
from enum import Enum, auto
from typing import Optional


class JiraSearchQueryBuilder:
    class __QueryParts(Enum):
        PROJECT = auto()
        TYPE = auto()
        STATUS = auto()
        RESOLUTION_DATE = auto()
        LAST_MODIFIED = auto()
        TEAM = auto()
        TASK_IDS = auto()
        ORDER_BY = auto()

    def __init__(self,
                 projects: list[str] = None,
                 statuses: list[str] = None,
                 task_types: list[str] = None,
                 teams: list[str] = None,
                 resolution_dates: tuple[Optional[datetime.datetime], Optional[datetime.datetime]] = None,
                 last_modified_dates: tuple[Optional[datetime.datetime], Optional[datetime.datetime]] = None,
                 task_ids: list[str] = None,
                 raw_queries: list[str] = None,
                 order_by: Optional[str] = None
                 ) -> None:
        self.query_parts = {}
        self.raw_queries: list[str] = []

        self.with_projects(projects)
        self.with_statuses(statuses)
        self.with_resolution_dates(resolution_dates)
        self.with_task_types(task_types)
        self.with_teams(teams)
        self.with_last_modified_dates(last_modified_dates)
        self.with_task_ids(task_ids)
        self.with_raw_queries(raw_queries)
        self.with_order_by(order_by)

    def with_projects(self, projects: list[str]):
        if projects is None:
            return
        project_filter = "project IN (" + ",".join(projects) + ")"
        self.__add_filter(self.__QueryParts.PROJECT, project_filter)

    def with_statuses(self, statuses):
        if statuses is None:
            return
        status_filter = "status in (" + self.__convert_in_jql_value_list(statuses) + ")"
        self.__add_filter(self.__QueryParts.STATUS, status_filter)

    def with_resolution_dates(self, resolution_dates: tuple[Optional[datetime.datetime], Optional[datetime.datetime]]):
        if resolution_dates is None:
            return
        date_filter = self.__create_date_range_filter("resolutiondate",
                                                      resolution_dates[0],
                                                      resolution_dates[1])
        if date_filter:
            self.__add_filter(self.__QueryParts.RESOLUTION_DATE, date_filter)

    def with_last_modified_dates(self, last_modified_datas: tuple[Optional[datetime.datetime], Optional[datetime.datetime]]):
        if last_modified_datas is None:
            return
        date_filter = self.__create_date_range_filter("updated",
                                                      last_modified_datas[0],
                                                      last_modified_datas[1])
        if date_filter:
            self.__add_filter(self.__QueryParts.LAST_MODIFIED, date_filter)

    def with_task_types(self, task_types):
        if task_types is None:
            return
        task_type_filter = "issuetype in (" + self.__convert_in_jql_value_list(task_types) + ")"
        self.__add_filter(self.__QueryParts.TYPE, task_type_filter)

    def with_task_ids(self, task_ids: list[str]):
        if task_ids is None:
            return
        ids_filter = "key in (" + ", ".join(task_ids) + ")"
        self.__add_filter(self.__QueryParts.TASK_IDS, ids_filter)

    def with_teams(self, teams: list[str]):
        if teams is None:
            return
        team_filter = "Team[Team] in (" + self.__convert_in_jql_value_list(teams) + ")"
        self.__add_filter(self.__QueryParts.TEAM, team_filter)

    def with_raw_queries(self, raw_queries: list[str]):
        if raw_queries is None:
            return
        normalized = [q.strip() for q in raw_queries if q and q.strip()]
        if not normalized:
            return
        self.raw_queries.extend(normalized)

    def with_order_by(self, order_by: str):
        if not order_by:
            return
        self.__add_filter(self.__QueryParts.ORDER_BY, order_by)

    def build_query(self) -> str:
        where_parts = [v for k, v in self.query_parts.items() if k != self.__QueryParts.ORDER_BY]
        if self.raw_queries:
            where_parts.extend([q.strip() for q in self.raw_queries if q and q.strip()])
        base = ' AND '.join(where_parts)
        order_by = self.query_parts.get(self.__QueryParts.ORDER_BY)
        if order_by:
            if base:
                return base + ' ORDER BY ' + order_by
            else:
                return 'ORDER BY ' + order_by
        return base

    @staticmethod
    def __convert_in_jql_value_list(statuses):
        return ', '.join(['"%s"' % w for w in statuses])

    @staticmethod
    def __create_date_range_filter(field_name: str, start_date: Optional[datetime.date], end_date: Optional[datetime.date]):
        parts = []
        if start_date is not None:
            start_date_str = start_date.strftime('%Y-%m-%d')
            parts.append(f"{field_name} >= '{start_date_str}'")
        if end_date is not None:
            end_date_str = end_date.strftime('%Y-%m-%d')
            parts.append(f"{field_name} <= '{end_date_str}'")
        return ' and '.join(parts)

    def __add_filter(self, query_part_type: __QueryParts, query_part):
        self.query_parts[query_part_type] = query_part.strip()




from typing import Any


def get_attribute_by_path(obj: Any, path: str, default: Any = None) -> Any:
    cur = obj
    if not path:
        return cur
    for part in path.split('.'):
        if cur is None:
            return default
        try:
            cur = getattr(cur, part)
        except Exception:
            return default
    return cur
